#include <unistd.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <stdio.h>
#include <string.h>

#define COLORMASK 0xee

#define BLACKLO   0x01
#define BLACKHI   0x10
#define CYANLO    0x08
#define CYANHI    0x80
#define MAGENTALO 0x04
#define MAGENTAHI 0x40
#define YELLOWLO  0x02
#define YELLOWHI  0x20 

long make_plane(long height,long width,char *inb,char *outb,
		int masklo,int maskhi,int shiftoff,long nibble){
  long j;
  long outbit=7;
  long outbyte=0;
  
  memset(outb,0,width/8+1);
    
  for(j=0;j<width;j++){
    int bit;
    if(nibble&1)
      bit=inb[(nibble>>1)]&masklo;
    else
      bit=(inb[(nibble>>1)]&maskhi)>>4;
    nibble++;
    outb[outbyte]|=((bit>>shiftoff)<<outbit);
    
    outbit--;
    if(outbit<0){
      outbit=7;
      outbyte++;
    }
  }
  return nibble;
}

int main(int argc,char *argv[]){
  long width,height,size,i;
  int mapfd,color=0;
  char *inb,*outb,*filename;
  if(!argv[1]){
    fprintf(stderr,"width (in dots) required");
    exit(1);
  }
  if(!argv[2]){
    fprintf(stderr,"height (in dots) required");
    exit(1);
  }
  
  width=atoi(argv[1]);
  height=atoi(argv[2]);
  size=width*height;

  /* We have to suck in the whole thing in order to check for
     non-black pixels.  In nibble-per-pixel format, that's only 15M
     for letter size.  Still gross, but an order of magnitude better
     than before. */

  filename=strdup("/tmp/bitcmyk2pbm_XXXXXX");
  mapfd=mkstemp(filename);
  unlink(filename);
  if(!mapfd){
    fprintf(stderr,"could not open temp file)");
    exit(1);
  }
  outb=inb=mmap(0,size+width/2+1,PROT_READ|PROT_WRITE,MAP_PRIVATE,mapfd,0);
  inb+=width/2+1;

  while(1){
    /* suck it down */
    long bytes=fread(inb,1,size/2,stdin);
    height=bytes*2/width;
    if(height<=0)break;

    /* look for color */
    for(i=0;i<size/2;i++)
      if(inb[i]&COLORMASK){
	color=1;
	break;
      }
    
    if(!color){
      long nibble=0;
      fprintf(stdout,"P4\n# generated by bitcmyk2pbm\n%ld %ld\n",width,height);
      
      for(i=0;i<height;i++){
	nibble=make_plane(height,width,inb,outb,BLACKLO,BLACKHI,0,nibble);
	fwrite(outb,1,(width+7)/8,stdout);
      }
    }else{
      long nibble=0;
      
      fprintf(stdout,"P4\n#PI%d %ld %ld\n%ld %ld\n",
	      4,width,height,width,height*4);
      
      for(i=0;i<height;i++){
	int l;
	make_plane(height,width,inb,outb,CYANLO,CYANHI,3,nibble);
	fwrite(outb,1,(width+7)/8,stdout);
	make_plane(height,width,inb,outb,MAGENTALO,MAGENTAHI,2,nibble);
	fwrite(outb,1,(width+7)/8,stdout);
	make_plane(height,width,inb,outb,YELLOWLO,YELLOWHI,1,nibble);
	fwrite(outb,1,(width+7)/8,stdout);
	nibble=make_plane(height,width,inb,outb,BLACKLO,BLACKHI,0,nibble);
	fwrite(outb,1,(width+7)/8,stdout);
      }
    }
  }
  munmap(outb,size+width/2);
  return(0);
}

