#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "misc.h"
#include "z51.h"

static void paper_shift(FILE *out,int offset){
  char *buf=alloca(sizeof(LEX_vshift));
  memcpy(buf,LEX_vshift,sizeof(buf));
  buf[3]=(char)(offset >> 8);
  buf[4]=(char)(offset & 0xFF);
  fwrite(buf,1,sizeof(LEX_vshift),out);
}

static inline int line_empty_p(char *buf,long len){
  return(buf[0]==0 && memcmp(buf,buf+1,len-1)==0);
}

/* return coordinate of leftmost pixel (in pixels) */
static long leftmost_pixel(char *buf, int bytelen){
  char *r=buf;
  
  /* don't think too hard.  Closest byte is fine */
  while(*r++==0 && --bytelen>0);
  return(r-buf-1)*8;
}

/* return coordinate of rightmost pixel (in pixels) */
static long rightmost_pixel(char *buf, int bytelen){
  char *r=buf+bytelen;

  /* don't think too hard.  Closest byte is fine */
  while(*--r==0 && --bytelen>0);
  return(r-buf+1)*8;
}

/* find leftmost and rightmost pixel in whole pass
 * buf        - pixel buffer
 * bytelen    - width of line (in bytes)
 * bufheight  - buffer height (pixels)
 * leftmost,
 *  rightmost - output
 */

static void find_lr_pixels(hres_settings *hm,
			   char *buf[],int bytelen,int bufheight,
			   long *leftmost, long *rightmost){
  long maxright=bytelen*8-1;
  long left=maxright;
  long right=0;
  long i;

  for(i=0;i<bufheight;i++){
    int ltmp,rtmp;

    if (line_empty_p(buf[i],bytelen))continue;
    
    ltmp=leftmost_pixel(buf[i],bytelen);
    rtmp=rightmost_pixel(buf[i],bytelen);
    left=(ltmp<left?ltmp:left);
    right=(rtmp>right?rtmp:right);
  }

  /* don't think too hard. Always on because we double lines for 300vdpi. */
  if(left<maxright){
    long intershift=MAX(hm->C_offset_headsep,hm->k_offset_headsep);
    left=MAX(left-intershift,0);
    /*right=MIN(left+intershift,maxright); Uhh.... this is unneccessary */
  }

  *leftmost  = left;
  *rightmost = right; 
  return;
}

/* RLE packet:
   every packet contains 13 info bits (12 for color, but this is juat a 
        special case of the 13 bit packet witht he 13 bit always 0) 
   1 = 8x left white, 8x right white
   0 = 2 data bytes follow (8 bits left, 8 bits right)
*/

/* for color packets, make sure to call with a 13 element vbuf and
   vbuf[12] set to 0 and rle2 set to 1 (is 0 okay? too many bits) */
int rle_packet(unsigned char *packetbuf,unsigned int *vbuf,int rlemode){
  unsigned int lastword=0,k;
  unsigned char *p=packetbuf+2; /* leave space for rle bytes */
  int bits=0;

  if(rlemode==2)lastword=0x8FFF; /* impossible value */  
  
  for(k=0;k<MAX_JET_WORDS;k++){
    unsigned int t=vbuf[k];
    if (t==lastword && rlemode!=0) 
      bits|=(1<<k); /* packet is empty */
    else{
      *(p++)=(t >> 8) & 0xFF;
      *(p++)=(t & 0xFF);
      if(rlemode==2)lastword=t;
    }
  }
  packetbuf[1]=(unsigned char)( (bits) & 0xFF);
  packetbuf[0]=(unsigned char)( ((bits>>8) & 0x1f));
  if(rlemode<2)packetbuf[0]|=0x20;

  return(p-packetbuf);
}

/* most important print packet
 * see lexmarkprotocol.txt for more information */
/*          length of complete sequence ---  vvvvvvvvv */
/* this is template now. Outbuf is allocated and initialized
 * by lxmgen_print_page() to avoid large static array
 * in gs binary
 */

static int print_cols(FILE *out,char *outbuf,
		      int left,int right,int vstart, int vend,
		      char *buf[],int width,
		      int margin,int headsep,int model,int color){

  char *p=outbuf+IDX_DATA;
  int i,k;

  outbuf[IDX_PACKETS]  =(unsigned char)((right-left) >> 8);
  outbuf[IDX_PACKETS+1]=(unsigned char)((right-left) & 0xFF);
  outbuf[IDX_HSTART]   =(unsigned char)((left+margin) >> 8);
  outbuf[IDX_HSTART+1] =(unsigned char)((left+margin) & 0xFF);
  outbuf[IDX_HEND]     =(unsigned char)((right+margin) >> 8);
  outbuf[IDX_HEND+1]   =(unsigned char)((right+margin) & 0xFF);
  
  if (model==LX_5700)
    outbuf[IDX_5700DIF]=(unsigned char) (color ? 0x11 : 0x1);
  outbuf[IDX_CARTRIDGE]= (unsigned char)(color ? 0 : 1 );

  /* fill columns */
  for(i=left;i<right;i++){
    unsigned int vbuf[MAX_JET_WORDS];
    unsigned char *vp,*vp2;
    int smask=0x80>>(i&0x7);
    int smask2=0x80>>((i+headsep)&0x7);
    long vi=vstart*2; /* vbuf index in pixels (not array index) */

    memset(vbuf,0,sizeof(vbuf));    
    
    for(k=vstart;k<vend;k+=2){
      vp=buf[k]+i/8; /* left inkjets */ 
      vp2=buf[k+1]+(i+headsep)/8; /* right inkjets */
      
      if (vp[0]&smask) 
	vbuf[vi/16] |= (0x8000>>(vi&0xf));
      vi++;
      if (i+headsep<width*8 && vp2[0]&smask2)
	vbuf[vi/16] |= (0x8000>>(vi&0xf));
      vi++;
    }

    /* what packet encoding?  rle1 or rle2? */
    {
      unsigned char rle1_packet[MAX_PACKET_SIZE];
      unsigned char rle2_packet[MAX_PACKET_SIZE];
     
      int rle1_size=rle_packet(rle1_packet,vbuf,1);
      int rle2_size=rle_packet(rle2_packet,vbuf,2);

      /* possible that ||p->color is needed */
      if(rle1_size>rle2_size){
	memcpy(p,rle2_packet,rle2_size);
	p+=rle2_size;
      }else{
	memcpy(p,rle1_packet,rle1_size);
	p+=rle1_size;
      }
    }
  }

  /* ------------ */
  {
    long clen=p-outbuf; 
    outbuf[IDX_SEQLEN-1]  =(unsigned char)(clen >> 16);
    outbuf[IDX_SEQLEN]  =(unsigned char)(clen >> 8);
    outbuf[IDX_SEQLEN+1]=(unsigned char)(clen & 0xFF);
    fwrite(outbuf,1,clen,out);
  }

  return 0;
}

typedef struct {
  long cpos_unprinted;
  long mpos_unprinted;
  long ypos_unprinted;
  long kpos_unprinted;
} interlace_field;  

/* Lexmark generic print page routine */
static int pages=0;
int lxmgen_print_page(long pwidth, long pheight, long bwidth,
		      char *cbuf,char *mbuf,char *ybuf, char *kbuf,
		      FILE *prn_stream,
		      hres_settings *hm,vres_settings *vm,int model){
  int i,j,k;
  char *outbuf,*zeroes;  
  char **swipeptr=alloca(MAX_SWIPE_H*sizeof(char *)); /* pointers to
							 buffer lines,
							 preparing for
							 interlacing */

  /* each color is handled in its own framing, but a color swipe will
     trigger and buffer printing for C, M and Y, buffering zeroes if
     that color has nothing yet to print.

     Black is done totally async in its own swipe 208 the full pixels
     high */

  /* as we move onto the paper, yellow breaks ground first, followed
     by m then c and then k */

  long master=-vm->top_offset;
  interlace_field even;
  interlace_field odd;
  interlace_field *current;
  long advlines=0;

  /* actual head y location */
  long cposition=master+vm->c_offset_y;
  long mposition=master+vm->m_offset_y;
  long yposition=master+vm->y_offset_y;
  long kposition=master+vm->k_offset_y;

  even.cpos_unprinted=-1;
  even.mpos_unprinted=-1;
  even.ypos_unprinted=-1;
  even.kpos_unprinted=-1;
  odd.cpos_unprinted=-1;
  odd.mpos_unprinted=-1;
  odd.ypos_unprinted=-1;
  odd.kpos_unprinted=-1;

  outbuf = calloc(OUT_BUF_SIZE,1);
  zeroes = calloc(bwidth,1);
  if (outbuf == NULL){
    fprintf(stderr,"malloc of %d bytes failed\n",OUT_BUF_SIZE);
    exit(1);
  }

  /* header */
  memcpy(outbuf,LEX_graphics,sizeof(LEX_graphics));

  /* adjust horizontal motor speed for current resolution */
  {
    char mspeed[3]={1,2,5}; /* LXR_300=0, LXR_600=1, LXR_1200=2 */
    outbuf[IDX_HORRES]=mspeed[hm->hres];
  }
  
  if(!pages)fwrite(LEX_fullinit,1,sizeof(LEX_fullinit),prn_stream);
  
  pages++;

  if(!cbuf)even.cpos_unprinted=pheight;
  if(!mbuf)even.mpos_unprinted=pheight;
  if(!ybuf)even.ypos_unprinted=pheight;
  if(!kbuf)even.kpos_unprinted=pheight;
  if(!cbuf)odd.cpos_unprinted=pheight;
  if(!mbuf)odd.mpos_unprinted=pheight;
  if(!ybuf)odd.ypos_unprinted=pheight;
  if(!kbuf)odd.kpos_unprinted=pheight;

  /* take it one line at a time */
  while(cposition<pheight && mposition<pheight && 
	yposition<pheight && kposition<pheight){
    int blacktrigger=0;
    int colortrigger=0;

    current=&even;
    if(vm->field_interlace && master&1)
      current=&odd;

    /* keep track of the next line to print if we don't already know it */
    if(cposition>=0 && current->cpos_unprinted==-1 && 
       !line_empty_p(cbuf+bwidth*cposition,bwidth))
      current->cpos_unprinted=cposition;

    if(mposition>=0 && current->mpos_unprinted==-1 && 
       !line_empty_p(mbuf+bwidth*mposition,bwidth))
      current->mpos_unprinted=mposition;

    if(yposition>=0 && current->ypos_unprinted==-1 && 
       !line_empty_p(ybuf+bwidth*yposition,bwidth))
      current->ypos_unprinted=yposition;

    if(kposition>=0 && current->kpos_unprinted==-1 && 
       !line_empty_p(kbuf+bwidth*kposition,bwidth))
      current->kpos_unprinted=kposition;

    /* do this now because we want these pointing one past the end */
    cposition++;
    mposition++;
    yposition++;
    kposition++;

    /* to the nitty gritty.  Does black trigger? */
    if(current->kpos_unprinted!=-1 &&
       kposition-current->kpos_unprinted>=vm->k_swipe_height)blacktrigger=1;
    /* Does any color trigger? */
    if(current->cpos_unprinted!=-1 &&
       cposition-current->cpos_unprinted>=vm->c_swipe_height)colortrigger=1;
    if(current->mpos_unprinted!=-1 &&
       mposition-current->mpos_unprinted>=vm->m_swipe_height)colortrigger=1;
    if(current->ypos_unprinted!=-1 &&
       yposition-current->ypos_unprinted>=vm->y_swipe_height)colortrigger=1;

    /* interlacing fudge; we want to stagger the fields, not have them
       bunch up at the same seam */
    if(vm->field_interlace){
      if(even.kpos_unprinted!=-1 && 
	 odd.kpos_unprinted!=-1 &&
	 labs(even.kpos_unprinted-odd.kpos_unprinted)<vm->k_swipe_height/3 &&
	 kposition-current->kpos_unprinted>=vm->k_swipe_height/2)
	blacktrigger=1;

      if(even.cpos_unprinted!=-1 && 
	 odd.cpos_unprinted!=-1 &&
	 labs(even.cpos_unprinted-odd.cpos_unprinted)<vm->c_swipe_height/3 &&
	 cposition-current->cpos_unprinted>=vm->c_swipe_height/2)
	colortrigger=1;
      
      if(even.mpos_unprinted!=-1 && 
	 odd.mpos_unprinted!=-1 &&
	 labs(even.mpos_unprinted-odd.mpos_unprinted)<vm->m_swipe_height/3 &&
	 mposition-current->mpos_unprinted>=vm->m_swipe_height/2)
	colortrigger=1;
      
      if(even.ypos_unprinted!=-1 && 
	 odd.ypos_unprinted!=-1 &&
	 labs(even.ypos_unprinted-odd.ypos_unprinted)<vm->y_swipe_height/3 &&
	 yposition-current->ypos_unprinted>=vm->y_swipe_height/2)
	colortrigger=1;
    }

    if(blacktrigger || colortrigger){
      if(advlines>0){
	int mult[3]={4,2,1}; /* multiply: 300dpi=4, 600dpi=2, 1200dpi=1 */
	paper_shift(prn_stream,advlines*mult[vm->vres]);
      }
      advlines=0;
    }

    if(blacktrigger){
      long leftmost, rightmost;

      /* make a full 208 physical pixel high black swipe */
      
      /* for 300 dpi, we assign a line to two inkjets (left and right of
	              consecutive lines) */
      /* for 600 dpi, we assign each line in order to an inkjet (L and R) */
      /* for 1200 dpi, we assign every other line to an inkjet (L and R) 
                       and make a second pass later to interlace the rest. 
                       Where 300dpi has to lay down extra ink, we want to
                       take ink away; We do this with a horizontal adjacency
                       elimination */
 
      for(i=0;i<MAX_SWIPE_H;i++)swipeptr[i]=zeroes;
      for(i=0,j=kposition-vm->k_swipe_height,k=K_BUFFER_VOFFSET;
	  i<vm->k_swipe_height;){
	if(j>=current->kpos_unprinted && j<pheight){
	  swipeptr[k]=kbuf+j*bwidth;
	  if(vm->field_nodouble==0)
	    swipeptr[k+1]=kbuf+j*bwidth;
	}
	k+=(2-vm->field_nodouble);
	j+=vm->field_interlace+1;
	i+=vm->field_interlace+1;
      }

      /* look for leftmost and rightmost pixels */
      find_lr_pixels(hm,swipeptr,bwidth,MAX_SWIPE_H,
		     &leftmost,&rightmost);
      if(leftmost<rightmost)
	print_cols(prn_stream,outbuf,
		   leftmost,rightmost,
		   0,MAX_SWIPE_H,
		   swipeptr,bwidth,
		   -hm->left_offset,
		   hm->k_offset_headsep,
		   model,
		   0);

      current->kpos_unprinted=-1;
    }

    if(colortrigger){
      long leftmost, rightmost;
      
      /* make a full 208 physical pixel high (192 actual color pixels)
	 color swipe of unprinted color areas */
      
      /* for 300 dpi, we assign a line to every other inkjet (just left) */
      /* for 600 dpi, we assign each line in order to an inkjet (L and R) */
      /* for 1200 dpi, we assign every other line to an inkjet (L and R) 
                       and make a second pass later to interlace the rest */
 
      for(i=0;i<MAX_SWIPE_H;i++)swipeptr[i]=zeroes;

      /* do we have cyan to print? */
      if(current->cpos_unprinted!=-1)
	for(i=0,j=cposition-vm->c_swipe_height,k=C_BUFFER_VOFFSET;
	    i<vm->c_swipe_height;){
	  if(j>=current->cpos_unprinted && j<pheight){
	    swipeptr[k]=cbuf+j*bwidth;

	    if(vm->field_nodouble==0)
	      swipeptr[k+1]=cbuf+j*bwidth;
	  }
	  k+=(2-vm->field_nodouble);
	  j+=vm->field_interlace+1;
	  i+=vm->field_interlace+1;
	}
      /* do we have magenta to print? */
      if(current->mpos_unprinted!=-1)
	for(i=0,j=mposition-vm->m_swipe_height,k=M_BUFFER_VOFFSET;
	    i<vm->m_swipe_height;){
	  if(j>=current->mpos_unprinted && j<pheight){
	    swipeptr[k]=mbuf+j*bwidth;
	    if(vm->field_nodouble==0)
	      swipeptr[k+1]=mbuf+j*bwidth;
	  }
	  k+=(2-vm->field_nodouble);
	  j+=vm->field_interlace+1;
	  i+=vm->field_interlace+1;
	}
      /* do we have yellow to print? */
      if(current->ypos_unprinted!=-1)
	for(i=0,j=yposition-vm->y_swipe_height,k=Y_BUFFER_VOFFSET;
	    i<vm->y_swipe_height;){
	  if(j>=current->ypos_unprinted && j<pheight){
	    swipeptr[k]=ybuf+j*bwidth;
	    if(vm->field_nodouble==0)
	      swipeptr[k+1]=ybuf+j*bwidth;
	  }
	  k+=(2-vm->field_nodouble);
	  j+=vm->field_interlace+1;
	  i+=vm->field_interlace+1;
	}

      /* look for leftmost and rightmost pixels */
      find_lr_pixels(hm,swipeptr,bwidth,MAX_SWIPE_H,
		     &leftmost,&rightmost);
      if(leftmost<rightmost)
	print_cols(prn_stream,outbuf,
		   leftmost,rightmost,
		   0,MAX_SWIPE_H,
		   swipeptr,bwidth,
		   -hm->left_offset,
		   hm->C_offset_headsep,model,1);
      
      if(cbuf)current->cpos_unprinted=-1;
      if(mbuf)current->mpos_unprinted=-1;
      if(ybuf)current->ypos_unprinted=-1;
    }
    advlines++;
    master++;
  }

  /* eject page */
  fwrite(LEX_eject,1,sizeof(LEX_eject),prn_stream);
  free(outbuf);
  free(zeroes);

  return 0;
}
